"""Objects representing a database schema."""

from __future__ import annotations

from typing import Any
from uuid import UUID

from pydantic import Field, SerializeAsAny, field_validator

from ultimate_notion.obj_api.core import GenericObject, TypedObject
from ultimate_notion.obj_api.enums import AggFunc, NumberFormat
from ultimate_notion.obj_api.objects import SelectGroup, SelectOption


class PropertyType(TypedObject, polymorphic_base=True):
    """Base class for Notion property objects."""

    id: str | None = None  # noqa: A003
    name: str | None = None

    @classmethod
    def build(cls):
        """Build the property value from given value, e.g. native Python or nested type.

        In practice, this is like calling __init__ with the corresponding keyword.
        """
        return cls.model_construct()


class Title(PropertyType, type='title'):
    """Defines the title configuration for a database property."""

    title: Any = Field(default_factory=dict)


class RichText(PropertyType, type='rich_text'):
    """Defines the rich text configuration for a database property."""

    rich_text: Any = Field(default_factory=dict)


class Number(PropertyType, type='number'):
    """Defines the number configuration for a database property."""

    class _NestedData(GenericObject):
        format: NumberFormat = NumberFormat.NUMBER  # noqa: A003

        # leads to better error messages, see
        # https://github.com/pydantic/pydantic/issues/355
        @field_validator('format')
        @classmethod
        def validate_enum_field(cls, field: str):
            return NumberFormat(field)

    number: _NestedData = _NestedData()

    @classmethod
    def build(cls, format):  # noqa: A002
        """Create a `Number` object with the expected format."""
        return cls.model_construct(number=cls._NestedData(format=format))


class Select(PropertyType, type='select'):
    """Defines the select configuration for a database property."""

    class _NestedData(GenericObject):
        options: list[SelectOption] = Field(default_factory=list)

    select: _NestedData = _NestedData()

    @classmethod
    def build(cls, options):
        """Create a `Select` object from the list of `SelectOption`'s."""
        return cls.model_construct(select=cls._NestedData(options=options))


class MultiSelect(PropertyType, type='multi_select'):
    """Defines the multi-select configuration for a database property."""

    class _NestedData(GenericObject):
        options: list[SelectOption] = Field(default_factory=list)

    multi_select: _NestedData = _NestedData()

    @classmethod
    def build(cls, options):
        """Create a `Select` object from the list of `SelectOption`'s."""
        return cls.model_construct(multi_select=cls._NestedData(options=options))


class Status(PropertyType, type='status'):
    """Defines the status configuration for a database property."""

    class _NestedData(GenericObject):
        options: list[SelectOption] = Field(default_factory=list)
        groups: list[SelectGroup] = Field(default_factory=list)

    status: _NestedData = _NestedData()


class Date(PropertyType, type='date'):
    """Defines the date configuration for a database property."""

    class _NestedData(GenericObject): ...

    date: _NestedData = _NestedData()


class People(PropertyType, type='people'):
    """Defines the people configuration for a database property."""

    class _NestedData(GenericObject): ...

    people: _NestedData = _NestedData()


class Files(PropertyType, type='files'):
    """Defines the files configuration for a database property."""

    class _NestedData(GenericObject): ...

    files: _NestedData = _NestedData()


class Checkbox(PropertyType, type='checkbox'):
    """Defines the checkbox configuration for a database property."""

    class _NestedData(GenericObject): ...

    checkbox: _NestedData = _NestedData()


class Email(PropertyType, type='email'):
    """Defines the email configuration for a database property."""

    class _NestedData(GenericObject): ...

    email: _NestedData = _NestedData()


class URL(PropertyType, type='url'):
    """Defines the URL configuration for a database property."""

    class _NestedData(GenericObject): ...

    url: _NestedData = _NestedData()


class PhoneNumber(PropertyType, type='phone_number'):
    """Defines the phone number configuration for a database property."""

    class _NestedData(GenericObject): ...

    phone_number: _NestedData = _NestedData()


class Formula(PropertyType, type='formula'):
    """Defines the formula configuration for a database property."""

    class _NestedData(GenericObject):
        expression: str = None  # type: ignore

    formula: _NestedData = _NestedData()

    @classmethod
    def build(cls, expression):
        return cls.model_construct(formula=cls._NestedData(expression=expression))


class PropertyRelation(TypedObject, polymorphic_base=True):
    """Defines common configuration for a property relation."""

    database_id: UUID = None  # type: ignore


class SinglePropertyRelation(PropertyRelation, type='single_property'):
    """Defines a one-way relation configuration for a database property."""

    single_property: Any = Field(default_factory=dict)

    @classmethod
    def build(cls, dbref):
        """Create a `single_property` relation using the target database reference.

        `dbref` must be either a string or UUID.
        """
        rel = SinglePropertyRelation.model_construct(database_id=dbref)
        return Relation.model_construct(relation=rel)


class DualPropertyRelation(PropertyRelation, type='dual_property'):
    """Defines a two-way relation configuration for a database property.

    If a two-way relation property X relates to Y then the two-way relation property Y relates to X.
    """

    class _NestedData(GenericObject):
        synced_property_name: str | None = None
        synced_property_id: str | None = None

    dual_property: _NestedData = _NestedData()

    @classmethod
    def build(cls, dbref):
        """Create a `dual_property` relation using the target database reference.

        `dbref` must be either a string or UUID.
        """
        rel = DualPropertyRelation.model_construct(database_id=dbref)
        return Relation.model_construct(relation=rel)


class Relation(PropertyType, type='relation'):
    """Defines the relation configuration for a database property."""

    relation: SerializeAsAny[PropertyRelation]


class Rollup(PropertyType, type='rollup'):
    """Defines the rollup configuration for a database property."""

    class _NestedData(GenericObject):
        function: AggFunc = AggFunc.COUNT

        relation_property_name: str | None = None
        relation_property_id: str | None = None

        rollup_property_name: str | None = None
        rollup_property_id: str | None = None

        # leads to better error messages, see
        # https://github.com/pydantic/pydantic/issues/355
        @field_validator('function')
        @classmethod
        def validate_enum_field(cls, field: str):
            return AggFunc(field)

    rollup: _NestedData = _NestedData()

    @classmethod
    def build(cls, relation, property, function):  # noqa: A002
        return Rollup.model_construct(
            rollup=cls._NestedData(function=function, relation_property_name=relation, rollup_property_name=property)
        )


class CreatedTime(PropertyType, type='created_time'):
    """Defines the created-time configuration for a database property."""

    class _NestedData(GenericObject): ...

    created_time: _NestedData = _NestedData()


class CreatedBy(PropertyType, type='created_by'):
    """Defines the created-by configuration for a database property."""

    class _NestedData(GenericObject): ...

    created_by: _NestedData = _NestedData()


class LastEditedBy(PropertyType, type='last_edited_by'):
    """Defines the last-edited-by configuration for a database property."""

    class _NestedData(GenericObject): ...

    last_edited_by: _NestedData = _NestedData()


class LastEditedTime(PropertyType, type='last_edited_time'):
    """Defines the last-edited-time configuration for a database property."""

    class _NestedData(GenericObject): ...

    last_edited_time: _NestedData = _NestedData()


class UniqueID(PropertyType, type='unique_id'):
    """Unique ID database property."""

    class _NestedData(GenericObject):
        prefix: str | None = None

    unique_id: _NestedData = _NestedData()


class Verification(PropertyType, type='verification'):
    """Verfication database property of Wiki databases."""

    class _NestedData(GenericObject): ...

    verification: _NestedData = _NestedData()


class RenameProp(GenericObject):
    """Use to rename a property during a database update."""

    name: str



"""Wrapper for property values of pages"""

from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import date as dt_date
from datetime import datetime
from typing import Any, TypeAlias

from pydantic import SerializeAsAny, field_validator, model_serializer

from ultimate_notion.obj_api.core import GenericObject, NotionObject
from ultimate_notion.obj_api.enums import VState
from ultimate_notion.obj_api.objects import (
    DateRange,
    FileObject,
    ObjectReference,
    RichTextObject,
    TypedObject,
    User,
)
from ultimate_notion.obj_api.schema import AggFunc, SelectOption

#: Notion's complex `Date` type, which is either a date or without time or a date range of the former.
DateType: TypeAlias = dt_date | tuple[dt_date, dt_date]  # Note that date is the superclass of datetime!


class PropertyValue(TypedObject, polymorphic_base=True):
    """Base class for Notion property values."""

    id: str = None  # type: ignore  # noqa: A003

    @classmethod
    def build(cls, value):
        """Build the property value from given value, e.g. native Python or nested type.

        In practice, this is like calling __init__ with the corresponding keyword.
        """
        return cls.model_construct(**{cls.model_fields['type'].get_default(): value})


class Title(PropertyValue, type='title'):
    """Notion title type."""

    title: list[SerializeAsAny[RichTextObject]] = None  # type: ignore


class RichText(PropertyValue, type='rich_text'):
    """Notion rich text type."""

    rich_text: list[SerializeAsAny[RichTextObject]] = None  # type: ignore


class Number(PropertyValue, type='number'):
    """Simple number type."""

    number: float | int | None = None  # ToDo: Recheck if it should be int | float | None instead


class Checkbox(PropertyValue, type='checkbox'):
    """Simple checkbox type; represented as a boolean."""

    checkbox: bool | None = None


class Date(PropertyValue, type='date'):
    """Notion complex date type - may include timestamp and/or be a date range."""

    date: DateRange | None = None

    @classmethod
    def build(cls, start: dt_date, end: dt_date | None = None):
        """Create a new Date from the native values."""
        return cls.model_construct(date=DateRange(start=start, end=end))


class Status(PropertyValue, type='status'):
    """Notion status property."""

    status: SelectOption | None = None


class Select(PropertyValue, type='select'):
    """Notion select type."""

    select: SelectOption | None = None


class MultiSelect(PropertyValue, type='multi_select'):
    """Notion multi-select type."""

    multi_select: list[SelectOption] = None  # type: ignore


class People(PropertyValue, type='people'):
    """Notion people type."""

    people: list[User] = None  # type: ignore

    # Custom serializer as we receive various UserRef subtypes but need to pass
    # a proper UserRef to the Notion API. It's just so inconsistent!
    @model_serializer
    def serialize(self) -> dict[str, Any]:
        return {'people': [{'id': user.id, 'object': user.object} for user in self.people]}


class URL(PropertyValue, type='url'):
    """Notion URL type."""

    url: str | None = None


class Email(PropertyValue, type='email'):
    """Notion email type."""

    email: str | None = None


class PhoneNumber(PropertyValue, type='phone_number'):
    """Notion phone type."""

    phone_number: str | None = None


class Files(PropertyValue, type='files'):
    """Notion files type."""

    files: list[SerializeAsAny[FileObject]] = None  # type: ignore


class FormulaResult(TypedObject, ABC, polymorphic_base=True):
    """A Notion formula result.

    This object contains the result of the expression in the database properties.
    """

    @property
    @abstractmethod
    def value(self):
        """Return the result of this FormulaResult."""


class StringFormula(FormulaResult, type='string'):
    """A Notion string formula result."""

    string: str | None = None

    @property
    def value(self):
        return self.string


class NumberFormula(FormulaResult, type='number'):
    """A Notion number formula result."""

    number: float | int | None = None

    @property
    def value(self):
        return self.number


class DateFormula(FormulaResult, type='date'):
    """A Notion date formula result."""

    date: DateRange | None = None

    @property
    def value(self) -> None | DateType:
        if self.date is None:
            return None
        elif self.date.end is None:
            return self.date.start
        else:
            return self.date.start, self.date.end


class BooleanFormula(FormulaResult, type='boolean'):
    """A Notion boolean formula result."""

    boolean: bool | None = None

    @property
    def value(self):
        return self.boolean


class Formula(PropertyValue, type='formula'):
    """A Notion formula property value."""

    formula: FormulaResult | None = None


class Relation(PropertyValue, type='relation'):
    """A Notion relation property value."""

    relation: list[ObjectReference] = None  # type: ignore
    has_more: bool = False

    @classmethod
    def build(cls, pages):
        """Return a `Relation` property with the specified pages."""
        return cls.model_construct(relation=[ObjectReference.build(page) for page in pages])


class RollupObject(TypedObject, ABC, polymorphic_base=True):
    """A Notion rollup property value."""

    function: AggFunc | None = None

    @property
    @abstractmethod
    def value(self):
        """Return the native representation of this Rollup object."""


class RollupNumber(RollupObject, type='number'):
    """A Notion rollup number property value."""

    number: float | int | None = None

    @property
    def value(self) -> float | int | None:
        """Return the native representation of this Rollup object."""
        return self.number


class RollupDate(RollupObject, type='date'):
    """A Notion rollup date property value."""

    date: DateRange | None = None

    @property
    def value(self) -> DateType | None:
        if self.date is None:
            return None
        elif self.date.end is None:
            return self.date.start
        else:
            return self.date.start, self.date.end


class RollupArray(RollupObject, type='array'):
    """A Notion rollup array property value."""

    array: list[PropertyValue]

    @property
    def value(self) -> list[PropertyValue]:
        """Return the native representation of this Rollup object."""
        return self.array


class Rollup(PropertyValue, type='rollup'):
    """A Notion rollup property value."""

    rollup: RollupObject | None = None


class CreatedTime(PropertyValue, type='created_time'):
    """A Notion created-time property value."""

    created_time: datetime


class CreatedBy(PropertyValue, type='created_by'):
    """A Notion created-by property value."""

    created_by: SerializeAsAny[User]


class LastEditedTime(PropertyValue, type='last_edited_time'):
    """A Notion last-edited-time property value."""

    last_edited_time: datetime


class LastEditedBy(PropertyValue, type='last_edited_by'):
    """A Notion last-edited-by property value."""

    last_edited_by: SerializeAsAny[User]


class UniqueID(PropertyValue, type='unique_id'):
    """A Notion unique-id property value."""

    class _NestedData(GenericObject):
        number: int = 0
        prefix: str | None = None

    unique_id: _NestedData = _NestedData()


class Verification(PropertyValue, type='verification'):
    """A Notion verification property value."""

    class _NestedData(GenericObject):
        state: VState = VState.UNVERIFIED
        verified_by: SerializeAsAny[User] | None = None
        date: datetime | None = None

        # leads to better error messages, see
        # https://github.com/pydantic/pydantic/issues/355
        @field_validator('state')
        @classmethod
        def validate_enum_field(cls, field: str):
            return VState(field)

    verification: _NestedData = _NestedData()


class PropertyItem(NotionObject, PropertyValue, object='property_item'):
    """A `PropertyItem` returned by the Notion API.

    Basic property items have a similar schema to corresponding property values.  As a
    result, these items share the `PropertyValue` type definitions.

    This class provides a placeholder for parsing property items, however objects
    parse by this class will likely be `PropertyValue`'s instead.

    Notion-API: https://developers.notion.com/reference/property-item-object
    """

    id: str  # noqa: A003



"""View representing the result of a Query."""

from __future__ import annotations

from collections.abc import Callable
from html import escape as htmlescape
from typing import TYPE_CHECKING, Any, TypeAlias, TypeVar

import numpy as np
import pandas as pd
from tabulate import tabulate

from ultimate_notion.obj_api.query import DBQueryBuilder
from ultimate_notion.objects import File, RichText
from ultimate_notion.page import Page
from ultimate_notion.text import html_img
from ultimate_notion.utils import SList, deepcopy_with_sharing, find_index, find_indices, get_repr, is_notebook

if TYPE_CHECKING:
    from ultimate_notion.database import Database

ColType: TypeAlias = str | list[str]
T = TypeVar('T')


class Col(str):
    """Column in a view, behaving like a string with some special attributes."""

    # ToDo: Implement me, in order to allow Spark-like sorting/filtering, e.g. `Col(Name) == "Florian"` in View


class View:
    def __init__(self, database: Database, pages: list[Page], query: DBQueryBuilder):
        self.database = database
        self._query = query
        self._title_col = database.schema.get_title_prop().name
        self._columns = self._get_columns(self._title_col)
        self._pages: np.ndarray = np.array(pages)
        self.default_limit = 10

        self.reset()

    def reset(self) -> View:
        """Reset the view, i.e. remove filtering, index and sorting."""
        self._icon_name: str | None = None
        self._id_name: str | None = None
        self._index_name: str | None = None
        self._row_indices = np.arange(len(self._pages))
        self._col_indices = np.arange(len(self._columns))
        return self

    def clone(self) -> View:
        """Clone the current view."""
        return deepcopy_with_sharing(self, shared_attributes=['database', '_pages', '_query'])

    def _get_columns(self, title_col: str) -> np.ndarray:
        """Make sure title column is the first columns."""
        cols = list(self.database.schema.to_dict().keys())
        cols.insert(0, cols.pop(cols.index(title_col)))
        return np.array(cols)

    @property
    def columns(self) -> list[str]:
        """Columns of the database view aligned with the elements of a row."""
        cols = list(self._columns[self._col_indices])
        if self.has_icon:
            cols.insert(0, self._icon_name)
        if self.has_id:
            cols.insert(0, self._id_name)
        if self.has_index:
            cols.insert(0, self._index_name)
        return cols

    def get_page(self, idx: int) -> Page:
        """Retrieve a page by index of the view."""
        return self._pages[self._row_indices[idx]]

    def search_page(self, name: str | RichText) -> SList[Page]:
        """Retrieve a page from this view by name"""
        pages = [page for page in self.to_pages() if page.title == name]
        return SList(pages)

    def get_row(self, idx: int) -> tuple[Any, ...]:
        page = self.get_page(idx)
        row: list[Any] = []
        for col in self.columns:
            if col == self._title_col:
                row.append(page.title)
            elif col == self._id_name:
                row.append(page.id)
            elif col == self._index_name:
                row.append(idx)
            elif col == self._icon_name:
                row.append(page.icon)
            else:
                row.append(page.props[col])
        return tuple(row)

    def to_pages(self) -> list[Page]:
        """Convert the view to a simple list of pages."""
        return [self.get_page(idx) for idx in range(len(self))]

    def to_rows(self) -> list[tuple[Any, ...]]:
        return [self.get_row(idx) for idx in range(len(self))]

    def to_pandas(self) -> pd.DataFrame:
        """Convert the view to a pandas dataframe."""
        # remove index as pandas uses its own
        view = self.without_index() if self.has_index else self
        return pd.DataFrame(view.to_rows(), columns=view.columns)

    def _html_for_icon(self, rows: list[Any], cols: list[str]) -> list[Any]:
        # escape everything as we ask tabulate not to do it
        rows = [[htmlescape(elem) if isinstance(elem, str) else elem for elem in row] for row in rows]
        if (title_idx := find_index(self._icon_name, cols)) is None:
            return rows
        for idx, row in enumerate(rows):
            page = self.get_page(idx)
            icon = html_img(page.icon.url, size=1.2) if isinstance(page.icon, File) else page.icon
            row[title_idx] = icon
        return rows

    def as_table(self, tablefmt: str | None = None) -> str:
        """Return the view in a given string table format.

        Some table formats:

        - plain: no pseudographics
        - simple: Pandoc's simple table, i.e. only dashes to separate header from content
        - github: GitHub flavored Markdown
        - simple_grid: uses dashes & pipes to separate cells
        - html: standard html markup

        Find more table formats under: https://github.com/astanin/python-tabulate#table-format
        """
        rows = self.to_rows()
        cols = self.columns

        if tablefmt is None:
            tablefmt = 'html' if is_notebook() else 'simple'

        if tablefmt == 'html':
            if self.has_icon:
                rows = self._html_for_icon(rows, cols)
                html_str = str(tabulate(rows, headers=cols, tablefmt='unsafehtml'))  # str() as tabulate wraps the str
            else:
                html_str = str(tabulate(rows, headers=cols, tablefmt='html'))  # str() as tabulate wraps the str
            return html_str
        else:
            return tabulate(rows, headers=cols, tablefmt=tablefmt)

    def show(self, *, simple: bool | None = None):
        """Show the database as human-readable table."""
        if simple:
            tablefmt = 'simple'
        elif simple is None:
            tablefmt = 'html' if is_notebook() else 'simple'
        else:
            tablefmt = 'html'

        table_str = self.as_table(tablefmt=tablefmt)

        if is_notebook() and (tablefmt == 'html'):
            from IPython.display import display_html  # noqa: PLC0415

            display_html(table_str)
        else:
            print(table_str)  # noqa: T201

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by Jupyter Lab automatically to display this view."""
        if len(self) > self.default_limit:
            html_str = self.limit(self.default_limit).as_table(tablefmt='html')
            html_str = f'<div style="display: inline-block;text-align: center;">{html_str}&#8942;</div></div>\n'
            return html_str
        else:
            return self.as_table(tablefmt='html')

    def __repr__(self) -> str:
        return get_repr(self, desc=self.database.title)

    def __str__(self) -> str:
        return self.as_table()

    def __len__(self):
        return len(self._row_indices)

    @property
    def has_index(self) -> bool:
        return self._index_name is not None

    def with_index(self, name='index') -> View:
        """Add an index column to the view."""
        if self.has_index and name == self._index_name:
            return self

        if name in self.columns:
            msg = f"index '{name}' is already a column name"
            raise RuntimeError(msg)

        view = self.clone()
        view._index_name = name
        return view

    def without_index(self) -> View:
        """Remove index column from the view."""
        if not self.has_index:
            return self

        view = self.clone()
        view._index_name = None
        return view

    @property
    def has_icon(self) -> bool:
        return self._icon_name is not None

    def with_icon(self, name='icon') -> View:
        """Show icons in HTML output."""
        if self.has_icon and name == self._icon_name:
            return self

        view = self.clone()
        view._icon_name = name
        return view

    def without_icon(self) -> View:
        """Don't show icons in HTML output."""
        if not self.has_icon:
            return self

        view = self.clone()
        view._icon_name = None
        return view

    @property
    def has_id(self) -> bool:
        return self._id_name is not None

    def with_id(self, name: str = 'id') -> View:
        """Add an id column to the view."""
        if self.has_id and name == self._id_name:
            return self

        view = self.clone()
        view._id_name = name
        return view

    def without_id(self) -> View:
        """Remove id column from the view."""
        if not self.has_id:
            return self

        view = self.clone()
        view._id_name = None
        return view

    def head(self, num: int) -> View:
        """Keep only the first `num` elements in view."""
        view = self.clone()
        view._row_indices = view._row_indices[:num]
        return view

    def limit(self, num: int) -> View:
        """Alias for `head`"""
        return self.head(num)

    def tail(self, num: int) -> View:
        """Keep only the last `num` elements in view."""
        view = self.clone()
        view._row_indices = view._row_indices[-num:]
        return view

    def select(self, *cols: str) -> View:
        """Select columns for the view"""
        curr_cols = self._columns  # we only consider non-meta columns, e.g. no index, etc.
        if not_included := set(cols) - set(curr_cols):
            msg = f"Some columns, i.e. {', '.join(not_included)}, are not in view"
            raise RuntimeError(msg)

        view = self.clone()
        select_col_indices = find_indices(list(cols), curr_cols)
        view._col_indices = view._col_indices[select_col_indices]
        return view

    def apply(self, func: Callable[[Page], T]) -> list[T]:
        """Apply function to all pages in view.

        Args:
            func: function taking a Page as input
        """
        return [func(page) for page in self.to_pages()]

    def reverse(self) -> View:
        """Reverse the order of the rows."""
        view = self.clone()
        view._row_indices = view._row_indices[::-1]
        return view

    def sort(self):
        """Sort the view with respect to some columns."""
        # ToDo: Implement me
        raise NotImplementedError

    def filter(self):  # noqa: A003
        """Filter the view."""
        # ToDo: Implement me
        raise NotImplementedError

    def reload(self) -> View:
        """Reload all pages by re-executing the query that generated the view."""
        view = self.clone()
        view._pages = np.array(self.database._pages_from_query(query=self._query))
        return view



"""All definitions and enumerations"""

from __future__ import annotations

from enum import Enum


class VState(str, Enum):
    """Verification states for pages in wiki databases."""

    VERIFIED = 'verified'
    UNVERIFIED = 'unverified'


class Color(str, Enum):
    """Basic colors"""

    DEFAULT = 'default'
    GRAY = 'gray'
    BROWN = 'brown'
    ORANGE = 'orange'
    YELLOW = 'yellow'
    GREEN = 'green'
    BLUE = 'blue'
    PURPLE = 'purple'
    PINK = 'pink'
    RED = 'red'


class BGColor(str, Enum):
    """Background colors for most textual blocks, e.g. paragraphs, callouts, etc."""

    DEFAULT = 'default'
    GRAY = 'gray_background'
    BROWN = 'brown_background'
    ORANGE = 'orange_background'
    YELLOW = 'yellow_background'
    GREEN = 'green_background'
    BLUE = 'blue_background'
    PURPLE = 'purple_background'
    PINK = 'pink_background'
    RED = 'red_background'


class CodeLang(str, Enum):
    """Coding languages for code blocks."""

    ABAP = 'abap'
    ARDUINO = 'arduino'
    BASH = 'bash'
    BASIC = 'basic'
    C = 'c'
    CLOJURE = 'clojure'
    COFFEESCRIPT = 'coffeescript'
    CPP = 'c++'
    CSHARP = 'c#'
    CSS = 'css'
    DART = 'dart'
    DIFF = 'diff'
    DOCKER = 'docker'
    ELIXIR = 'elixir'
    ELM = 'elm'
    ERLANG = 'erlang'
    FLOW = 'flow'
    FORTRAN = 'fortran'
    FSHARP = 'f#'
    GHERKIN = 'gherkin'
    GLSL = 'glsl'
    GO = 'go'
    GRAPHQL = 'graphql'
    GROOVY = 'groovy'
    HASKELL = 'haskell'
    HTML = 'html'
    JAVA = 'java'
    JAVASCRIPT = 'javascript'
    JSON = 'json'
    JULIA = 'julia'
    KOTLIN = 'kotlin'
    LATEX = 'latex'
    LESS = 'less'
    LISP = 'lisp'
    LIVESCRIPT = 'livescript'
    LUA = 'lua'
    MAKEFILE = 'makefile'
    MARKDOWN = 'markdown'
    MARKUP = 'markup'
    MATLAB = 'matlab'
    MERMAID = 'mermaid'
    NIX = 'nix'
    OBJECTIVE_C = 'objective-c'
    OCAML = 'ocaml'
    PASCAL = 'pascal'
    PERL = 'perl'
    PHP = 'php'
    PLAIN_TEXT = 'plain text'
    POWERSHELL = 'powershell'
    PROLOG = 'prolog'
    PROTOBUF = 'protobuf'
    PYTHON = 'python'
    R = 'r'
    REASON = 'reason'
    RUBY = 'ruby'
    RUST = 'rust'
    SASS = 'sass'
    SCALA = 'scala'
    SCHEME = 'scheme'
    SCSS = 'scss'
    SHELL = 'shell'
    SQL = 'sql'
    SWIFT = 'swift'
    TYPESCRIPT = 'typescript'
    VB_NET = 'vb.net'
    VERILOG = 'verilog'
    VHDL = 'vhdl'
    VISUAL_BASIC = 'visual basic'
    WEBASSEMBLY = 'webassembly'
    XML = 'xml'
    YAML = 'yaml'
    MISC = 'java/c/c++/c#'


class AggFunc(str, Enum):
    """Aggregation functions for formulas."""

    COUNT = 'count'
    COUNT_VALUES = 'count_values'
    COUNT_PER_GROUP = 'count_per_group'

    EMPTY = 'empty'
    NOT_EMPTY = 'not_empty'

    CHECKED = 'checked'
    UNCHECKED = 'unchecked'

    PERCENT_EMPTY = 'percent_empty'
    PERCENT_NOT_EMPTY = 'percent_not_empty'
    PERCENT_CHECKED = 'percent_checked'
    PERCENT_PER_GROUP = 'percent_per_group'

    AVERAGE = 'average'
    MIN = 'min'
    MAX = 'max'
    MEDIAN = 'median'
    RANGE = 'range'
    SUM = 'sum'

    DATE_RANGE = 'date_range'
    EARLIEST_DATE = 'earliest_date'
    LATEST_DATE = 'latest_date'

    SHOW_ORIGINAL = 'show_original'
    SHOW_UNIQUE = 'show_unique'
    UNIQUE = 'unique'


class NumberFormat(str, Enum):
    """Number formats for numbers."""

    NUMBER = 'number'
    NUMBER_WITH_COMMAS = 'number_with_commas'
    PERCENT = 'percent'
    DOLLAR = 'dollar'
    CANADIAN_DOLLAR = 'canadian_dollar'
    EURO = 'euro'
    POUND = 'pound'
    YEN = 'yen'
    RUBLE = 'ruble'
    RUPEE = 'rupee'
    WON = 'won'
    YUAN = 'yuan'
    REAL = 'real'
    LIRA = 'lira'
    RUPIAH = 'rupiah'
    FRANC = 'franc'
    HONG_KONG_DOLLAR = 'hong_kong_dollar'
    NEW_ZEALAND_DOLLAR = 'new_zealand_dollar'
    KRONA = 'krona'
    NORWEGIAN_KRONE = 'norwegian_krone'
    MEXICAN_PESO = 'mexican_peso'
    RAND = 'rand'
    NEW_TAIWAN_DOLLAR = 'new_taiwan_dollar'
    DANISH_KRONE = 'danish_krone'
    ZLOTY = 'zloty'
    BAHT = 'baht'
    FORINT = 'forint'
    KORUNA = 'koruna'
    SHEKEL = 'shekel'
    CHILEAN_PESO = 'chilean_peso'
    PHILIPPINE_PESO = 'philippine_peso'
    DIRHAM = 'dirham'
    COLOMBIAN_PESO = 'colombian_peso'
    RIYAL = 'riyal'
    RINGGIT = 'ringgit'
    LEU = 'leu'
    ARGENTINE_PESO = 'argentine_peso'
    URUGUAYAN_PESO = 'uruguayan_peso'



"""Functionality to initialize a Notion Session."""

from __future__ import annotations

import logging
import os
from threading import RLock
from types import TracebackType
from typing import Any, ClassVar, cast
from uuid import UUID

import notion_client
from httpx import ConnectError
from notion_client.errors import APIResponseError

from ultimate_notion.blocks import Block, DataObject
from ultimate_notion.database import Database
from ultimate_notion.obj_api.endpoints import NotionAPI
from ultimate_notion.objects import RichText, User
from ultimate_notion.page import Page
from ultimate_notion.props import Title
from ultimate_notion.schema import DefaultSchema, PageSchema
from ultimate_notion.utils import ObjRef, SList, get_uuid

_log = logging.getLogger(__name__)
ENV_NOTION_TOKEN = 'NOTION_TOKEN'  # same as in `notion-sdk-py` package


class SessionError(Exception):
    """Raised when there are issues with the Notion session."""

    def __init__(self, message):
        """Initialize the `SessionError` with a supplied message."""
        super().__init__(message)


class Session:
    """A session for the Notion API.

    The session keeps tracks of all objects, e.g. pages, databases, etc.
    in an object store to avoid unnecessary calls to the API.
    """

    client: notion_client.Client
    api: NotionAPI
    _active_session: Session | None = None
    _lock = RLock()
    cache: ClassVar[dict[UUID, DataObject | User]] = {}

    def __init__(self, auth: str | None = None, **kwargs: Any):
        """Initialize the `Session` object and the raw `api` endpoints.

        Args:
            auth: secret token from the Notion integration
            **kwargs: Arguments for the [Notion SDK Client][https://ramnes.github.io/notion-sdk-py/reference/client/]
        """
        if auth is None:
            if (env_token := os.getenv(ENV_NOTION_TOKEN)) is not None:
                auth = env_token
            else:
                msg = f'Either pass `auth` or set {ENV_NOTION_TOKEN}'
                raise RuntimeError(msg)

        _log.info('Initializing Notion session...')
        Session._initialize_once(self)

        # Same sane default as notion_client defines its own logger
        kwargs.setdefault('logger', logging.getLogger(f'{__name__}.client'))
        kwargs.setdefault('log_level', logging.NOTSET)
        self.client = notion_client.Client(auth=auth, **kwargs)
        self.api = NotionAPI(self.client)

    @classmethod
    def _initialize_once(cls, instance: Session):
        with Session._lock:
            if Session._active_session and Session._active_session is not instance:
                msg = 'Cannot initialize multiple Sessions at once'
                raise ValueError(msg)
            else:
                Session._active_session = instance

    @classmethod
    def get_active(cls) -> Session:
        """Return the current active session or None."""
        with Session._lock:
            if Session._active_session:
                return Session._active_session
            else:
                msg = 'There is no activate Session'
                raise ValueError(msg)

    @classmethod
    def get_or_create(cls, *args, **kwargs) -> Session:
        """Return the current active session or create a new session."""
        with Session._lock:
            if Session._active_session:
                return Session._active_session
            else:
                return cls(*args, **kwargs)

    def __enter__(self) -> Session:
        # Do nothing here as `__init__` created a client already and calling
        # `self.client.__enter__()` would initialize another client internally
        return self

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        _log.info('Closing connection to Notion...')
        self.client.__exit__(exc_type, exc_value, traceback)
        Session._active_session = None
        Session.cache.clear()

    def close(self):
        """Close the session and release resources."""
        _log.info('Closing connection to Notion...')
        self.client.close()
        Session._active_session = None
        Session.cache.clear()

    def is_closed(self) -> bool:
        """Determine if the session is closed or not."""
        try:
            self.raise_for_status()
        except RuntimeError:  # used by httpx client in case of closed connection
            return True
        else:
            return False

    def raise_for_status(self):
        """Confirm that the session is active and raise otherwise.

        Raises SessionError if there is a problem, otherwise returns None.
        """
        try:
            self.whoami()
        except ConnectError as err:
            msg = 'Unable to connect to Notion'
            raise SessionError(msg) from err
        except APIResponseError as err:
            msg = 'Invalid API reponse'
            raise SessionError(msg) from err

    def create_db(self, parent: Page, schema: type[PageSchema] | None = None) -> Database:
        """Create a new database within a page.

        Implementation:

            1. initialize external forward relations, i.e. relations pointing to other databases
            2. create the database using a Notion API call and potential external forward relations
            3. initialize self-referencing forward relations
            4. create columns with self-referencing forward relations using an update call
            5. update the backward references, i.e. two-way relations, using an update call
        """
        if schema:
            schema._init_fwd_rels()
            schema_dct = {col.name: col.type.obj_ref for col in schema._get_init_cols()}
            title = schema.db_title.obj_ref if schema.db_title is not None else None
            db_obj = self.api.databases.create(parent=parent.obj_ref, title=title, schema=schema_dct)
            if schema.db_desc:
                db_obj = self.api.databases.update(db_obj, description=schema.db_desc.obj_ref)
        else:
            schema_dct = {col.name: col.type.obj_ref for col in DefaultSchema._get_init_cols()}
            db_obj = self.api.databases.create(parent=parent.obj_ref, schema=schema_dct)

        db: Database = Database.wrap_obj_ref(db_obj)

        if schema:
            db.schema = schema  # schema is thus bound
            schema._init_self_refs()
            schema._update_bwd_rels()

        self.cache[db.id] = db
        return db

    def create_dbs(self, parents: Page | list[Page], schemas: list[type[PageSchema]]) -> list[Database]:
        """Create new databases in the right order in case there a relations between them."""
        # ToDo: Implement
        raise NotImplementedError

    def get_or_create_db(self, parent: Page, schema: type[PageSchema], title: str | None = None):
        """Get or create the database."""
        # ToDo: Implement
        raise NotImplementedError

    def search_db(
        self, db_name: str | None = None, *, exact: bool = True, reverse: bool = False, deleted: bool = False
    ) -> SList[Database]:
        """Search a database by name or return all if `db_name` is None.

        Args:
            db_name: name/title of the database, return all if `None`
            exact: perform an exact search, not only a substring match
            reverse: search in the reverse order, i.e. the least recently edited results first
            deleted: include deleted databases in search
        """
        query = self.api.search(db_name).filter(db_only=True)
        if reverse:
            query.sort(ascending=True)
        dbs = [cast(Database, self.cache.setdefault(db.id, Database.wrap_obj_ref(db))) for db in query.execute()]
        if exact and db_name is not None:
            dbs = [db for db in dbs if db.title == db_name]
        if not deleted:
            dbs = [db for db in dbs if not db.is_deleted]
        return SList(dbs)

    def get_db(self, db_ref: ObjRef, *, use_cache: bool = True) -> Database:
        """Retrieve Notion database by uuid"""
        db_uuid = get_uuid(db_ref)
        if use_cache and db_uuid in self.cache:
            return cast(Database, self.cache[db_uuid])
        else:
            db = Database.wrap_obj_ref(self.api.databases.retrieve(db_uuid))
            self.cache[db.id] = db
            return db

    def search_page(
        self, title: str | None = None, *, exact: bool = True, reverse: bool = False, deleted: bool = False
    ) -> SList[Page]:
        """Search a page by name.

        Args:
            title: title of the page, return all if `None`
            exact: perform an exact search, not only a substring match
            reverse: search in the reverse order, i.e. the least recently edited results first
            deleted: include deleted pages in search
        """
        query = self.api.search(title).filter(page_only=True)
        if reverse:
            query.sort(ascending=True)
        pages = [
            cast(Page, self.cache.setdefault(page_obj.id, Page.wrap_obj_ref(page_obj))) for page_obj in query.execute()
        ]
        if exact and title is not None:
            pages = [page for page in pages if page.title == title]
        if not deleted:
            pages = [page for page in pages if not page.is_deleted]
        return SList(pages)

    def get_page(self, page_ref: ObjRef, *, use_cache: bool = True) -> Page:
        page_uuid = get_uuid(page_ref)
        if use_cache and page_uuid in self.cache:
            return cast(Page, self.cache[page_uuid])
        else:
            page = Page.wrap_obj_ref(self.api.pages.retrieve(page_uuid))
            self.cache[page.id] = page
            return page

    def create_page(self, parent: Page, title: RichText | str | None = None) -> Page:
        title_obj = title if title is None else Title(title).obj_ref
        page = Page.wrap_obj_ref(self.api.pages.create(parent=parent.obj_ref, title=title_obj))
        self.cache[page.id] = page
        return page

    def get_user(self, user_ref: ObjRef, *, use_cache: bool = True) -> User:
        """Get a user by id"""
        user_uuid = get_uuid(user_ref)
        if use_cache and user_uuid in self.cache:
            return cast(User, self.cache[user_uuid])
        else:
            user = User.wrap_obj_ref(self.api.users.retrieve(user_uuid))
            self.cache[user.id] = user
            return user

    def search_user(self, name: str) -> SList[User]:
        """Search a user by name"""
        return SList(user for user in self.all_users() if user.name == name)

    def whoami(self) -> User:
        """Return the user object of this bot."""
        user = self.api.users.me()
        return cast(User, self.cache.setdefault(user.id, User.wrap_obj_ref(user)))

    def all_users(self) -> list[User]:
        """Retrieve all users of this workspace."""
        return [
            cast(User, self.cache.setdefault(user.id, User.wrap_obj_ref(user))) for user in self.api.users.as_list()
        ]

    def _get_block(self, block_ref: ObjRef) -> Block:
        """Retrieve a single block by an object reference."""
        block_uuid = get_uuid(block_ref)
        return Block.wrap_obj_ref(self.api.blocks.retrieve(block_uuid))

    def _get_blocks(self, parent_ref: ObjRef) -> list[Block]:
        """Retrieve all blocks of a parent block"""
        parent_uuid = get_uuid(parent_ref)
        return [Block.wrap_obj_ref(block) for block in self.api.blocks.children.list(parent=parent_uuid)]
